import os

from signal import SIGINT

from copy import deepcopy

from typing import Dict, Optional, List, Callable
from abc import ABC, abstractmethod

from enum import IntEnum
from subprocess import Popen
from threading import Condition

from logging import Logger, INFO, ERROR

from .config import ProgramConfig


class ProgramObserver(ABC):
    @abstractmethod
    def get_name(self) -> str:
        pass

    @abstractmethod
    def spawn(
        self, name: str, logger: Logger, extra: Dict[str, str]
    ) -> "ProgramObserver":
        pass

    @abstractmethod
    def on_state_change(self, state: int):
        pass

    @abstractmethod
    def on_received_state_change(self, state: int):
        pass

    @abstractmethod
    def on_wait_timeout(self, elapsed_time: float):
        pass

    @abstractmethod
    def on_start(self):
        pass

    @abstractmethod
    def on_execution(self, cmd: List[str], env: Dict[str, str], config: ProgramConfig):
        pass

    @abstractmethod
    def on_wait(self, startup_time: float):
        pass

    @abstractmethod
    def on_run(self, pid: int):
        pass

    @abstractmethod
    def on_exit(self, rc: int):
        pass

    @abstractmethod
    def on_terminate(self, pid: int):
        pass

    @abstractmethod
    def on_interrupt(self, pid: int):
        pass

    @abstractmethod
    def on_stopped(self):
        pass

    @abstractmethod
    def on_cancel(self):
        pass

    @abstractmethod
    def on_crash(self, cmd: List[str], e: BaseException):
        pass


class LoggingProgramObserver(ProgramObserver):
    def __init__(self, name: str, logger: Logger, extra: Dict[str, str]) -> None:
        self.name = name
        self.logger = logger
        self.extra = extra

    def get_name(self) -> str:
        return self.name

    def spawn(
        self, name: str, logger: Logger, extra: Dict[str, str]
    ) -> ProgramObserver:
        return LoggingProgramObserver(name, logger, deepcopy(extra))

    def on_state_change(self, state: int):
        self.logger.debug("Changing state to %s", state, extra=self.extra)

    def on_received_state_change(self, state: int):
        self.logger.debug("Changing state to %s", str(state), extra=self.extra)

    def on_wait_timeout(self, elapsed_time: float):
        self.logger.debug(
            "Wait timed out after %f seconds", elapsed_time, extra=self.extra
        )

    def on_start(self):
        self.logger.debug("Waiting for the program to start...", extra=self.extra)

    def on_execution(self, cmd: List[str], env: Dict[str, str], config: ProgramConfig):
        self.logger.debug("Executing %s", str(cmd), extra=self.extra)
        self.logger.debug(
            "Environment %s",
            str(env),
            extra=self.extra,
        )

        self.logger.debug("Config: %s", str(config), extra=self.extra)

        if config.user:
            self.logger.debug("User id: %d", config.user, extra=self.extra)

        if config.umask:
            self.logger.debug(
                "umask: 0o%s", format(config.umask, "o"), extra=self.extra
            )

    def on_wait(self, startup_time: float):
        self.logger.debug(
            "Waiting %1.2f seconds to start...", startup_time, extra=self.extra
        )

    def on_run(self, pid: int):
        self.logger.debug("Process pid %d", pid, extra=self.extra)

    def on_exit(self, rc: int):
        self.logger.log(
            ERROR if rc else INFO, "Exited with rc: %d", rc, extra=self.extra
        )

    def on_terminate(self, pid: int):
        self.logger.debug("Terminating process %d", pid, extra=self.extra)

    def on_interrupt(self, pid: int):
        self.logger.debug("Interrupting process %d", pid, extra=self.extra)

    def on_stopped(self):
        self.logger.info("Program has stopped.", extra=self.extra)

    def on_cancel(self):
        self.logger.info("Program start canceled.", extra=self.extra)

    def on_crash(self, cmd: List[str], e: BaseException):
        self.logger.error(
            "Failed to execute command %s: %s", str(e), str(cmd), extra=self.extra
        )


class ProgramCanceledException(Exception):
    pass


class ProgramCrashedException(Exception):
    pass


class ProgramState(IntEnum):
    """
    State of a Process within a Program

    +------------+-------+----------------------------------------+
    | Symbol     | Value | Description                            |
    +============+=======+========================================+
    | INIT       | 0     | Process is initialized                 |
    +------------+-------+----------------------------------------+
    | WAITING    | 1     | Process is waiting to start            |
    +------------+-------+----------------------------------------+
    | STARTING   | 2     | Process is starting                    |
    +------------+-------+----------------------------------------+
    | TIMEOUT    | 3     | Process start timet out                |
    +------------+-------+----------------------------------------+
    | RUNNING    | 4     | Process is running                     |
    +------------+-------+----------------------------------------+
    | CANCELING  | 6     | The Process start is being canceled    |
    +------------+-------+----------------------------------------+
    | STOPPING   | 5     | The Process was terminated by encab    |
    +------------+-------+----------------------------------------+
    | CANCELED   | 6     | Process start is being canceled        |
    +------------+-------+----------------------------------------+
    | CRASHED    | 6     | Process crashed during start           |
    +------------+-------+----------------------------------------+
    | STOPPED    | 7     | The Process was terminated by encab    |
    +------------+-------+----------------------------------------+
    | FAILED     | 8     | Process ended with rc != 0             |
    +------------+-------+----------------------------------------+
    | SUCCEEDED  | 9     | Process succeeded with rc = 0          |
    +------------+-------+----------------------------------------+
    | CANCELED   | 10    | Process execution was canceled         |
    +------------+-------+----------------------------------------+
    """

    INIT = 0
    WAITING = 1
    STARTING = 2
    RUNNING = 3

    CANCELING = 4
    STOPPING = 5

    CANCELED = 6
    CRASHED = 7
    STOPPED = 8
    FAILED = 9
    SUCCEEDED = 10


class ProgramStateHandler(object):
    def __init__(self, observer: ProgramObserver) -> None:
        super().__init__()
        self._state = ProgramState.INIT
        self._cond = Condition()
        self._observer = observer

    def set(self, state: ProgramState) -> None:
        with self._cond:
            self._observer.on_state_change(state)
            self._state = state
            self._cond.notify_all()

    def get(self) -> ProgramState:
        with self._cond:
            return self._state

    def wait_for(
        self, condition: Callable[[int], bool], timeout: Optional[float] = None
    ) -> ProgramState:
        with self._cond:
            condition_met = self._cond.wait_for(lambda: condition(self._state), timeout)
            if not condition_met:
                self._observer.on_wait_timeout(timeout or -1)
            return self._state

    def wait_for_startup(self, timeout: Optional[float] = None) -> ProgramState:
        return self.wait_for(
            lambda state: state == ProgramState.WAITING
            or state >= ProgramState.RUNNING,
            timeout,
        )

    def join(self, timeout: Optional[float] = None) -> ProgramState:
        return self.wait_for(
            lambda state: state == ProgramState.WAITING
            or state >= ProgramState.CANCELED,
            timeout,
        )

    def join_wait(self, timeout: Optional[float] = None) -> ProgramState:
        return self.wait_for(
            lambda state: state >= ProgramState.CANCELED,
            timeout,
        )

    def wait(self, timeout: float = 0) -> None:
        def canceled():
            return self._state >= ProgramState.CANCELING

        with self._cond:
            if timeout == 0 or self._state >= ProgramState.STARTING:
                return

            self._state = ProgramState.WAITING
            self._cond.notify_all()

            self._observer.on_wait(timeout)
            was_canceled = self._cond.wait_for(canceled, timeout)

            if was_canceled:
                raise ProgramCanceledException()

    def handle_exit(self, exit_code: int, command: List[str]):
        if exit_code == 0:
            self._observer.on_exit(exit_code)
            self.set(ProgramState.SUCCEEDED)
            return

        state = self.get()
        if state == ProgramState.RUNNING:
            self._observer.on_exit(exit_code)
            self.set(ProgramState.FAILED)
        elif state == ProgramState.STOPPING:
            self._observer.on_stopped()
            self.set(ProgramState.STOPPED)
        else:
            self._observer.on_crash(
                command,
                ProgramCrashedException(f"State: {state}, exit code {exit_code}"),
            )
            self.set(ProgramState.CRASHED)

    def kill(self, process: Optional[Popen], signal):
        with self._cond:
            if self._state >= ProgramState.CANCELED:
                return

            if self._state == ProgramState.WAITING:
                self._state = ProgramState.CANCELING
            else:
                self._state = ProgramState.STOPPING
                pid = process.pid if process else None
                if pid:
                    if signal == SIGINT:
                        self._observer.on_interrupt(pid)
                    else:
                        self._observer.on_terminate(pid)

                    try:
                        os.kill(pid, signal)
                    except ProcessLookupError:
                        pass

            self._cond.notify_all()



from typing import Dict, Optional, List

from .config import ProgramConfig, EncabConfig, ConfigError
from .program import ExecutionContext, Program


class Programs(object):
    """
    controls the lifecycles of all configured programs
    """

    def __init__(
        self,
        program_configs: Dict[str, ProgramConfig],
        context: ExecutionContext,
        args: List[str],
        encab_config: Optional[EncabConfig] = None,
    ) -> None:
        """
        creates a programs instance and sets/replaces
        the main program form the arguments.

        :param program_configs: the program configurations
        :type program_configs: Dict[str, ProgramConfig]
        :param context: the root program context
        :type context: ExecutionContext
        :param args: the command line arguments
        :type args: List[str]
        :param encab_config: the basic encab config, defaults to None
        :type encab_config: Optional[EncabConfig], optional
        :raises ConfigError: If programs connot find a program to run
        """
        main: Optional[Program] = None
        self.helpers: List[Program] = list()

        for name, program_config in program_configs.items():
            if encab_config:
                program_config.extend(encab_config)
                program_config.set_user()

            if name == "main":
                if args:
                    program_config.command = args
                main = Program(name, program_config, context)
            else:
                self.helpers.append(Program(name, program_config, context))

        if not main and args:
            program_config = ProgramConfig.create(command=args, environment={})

            if encab_config:
                program_config.extend(encab_config)

            main = Program("main", program_config, context)

        if not main:
            raise ConfigError(
                "No main program to run! "
                "The main program must be defined in encab.yml or passed as argument on startup!"
            )

        self.main: Program = main

    def start_helpers(self):
        for helper in self.helpers:
            helper.start(helper.config.join_time)

    def stop_helpers(self):
        for helper in reversed(self.helpers):
            helper.terminate()

        for helper in reversed(self.helpers):
            helper.join(helper.config.join_time)

    def run(
        self,
    ):
        """
        runs the programs as configured and waits until the main program has stopped.
        """
        self.start_helpers()

        self.main.start()
        self.main.join_wait()

        self.stop_helpers()

    def start(
        self,
    ):
        """
        starts the programs as configured
        """

        self.start_helpers()
        self.main.start()

    def terminate(self):
        """
        terminates the main program and stops the helpers
        """
        self.main.terminate()
        self.stop_helpers()

    def interrupt(self):
        """
        interrupts the main program and stops the helpers
        """
        self.main.interrupt()
        self.stop_helpers()



import sys

from copy import deepcopy

from typing import Dict, Optional, IO, Union, List

from subprocess import Popen, PIPE
from threading import Thread
from signal import SIGINT, SIGTERM

from logging import Logger, INFO, ERROR, getLogger

from .config import ProgramConfig
from .program_state import (
    ProgramObserver,
    ProgramState,
    ProgramStateHandler,
    ProgramCanceledException,
)
from .extensions import extensions


class LogStream(object):
    """
    Reads from a stream in a background thread and loggs the result line by line
    """

    def __init__(
        self, logger: Logger, log_level: int, stream: IO[bytes], extra: Dict[str, str]
    ) -> None:
        """
        :param Logger logger: the logger to which the stream content is written
        :param int log_level: the log level (see Python logging)
        :param IOBase stream: the stream that is logged
        :param Dict[str, str] extra: extra information that is logged each line (see Python logging)
        """
        self.logger = logger
        self.log_level = log_level
        self.stream = stream
        self.extra = extra
        self.thread: Optional[Thread] = None

    def _run(self):
        try:
            for line in self.stream:
                strline = line.decode(sys.getdefaultencoding()).rstrip("\r\n\t ")
                self.logger.log(self.log_level, strline, extra=self.extra)
        except ValueError:
            pass  # stream was closed
        except OSError:
            self.logger.exception(
                "I/O Error while logging: %s", self.name, extra=self.extra  # type: ignore
            )
        except:
            self.logger.exception(
                "Something went wrong while logging", extra=self.extra
            )
            raise

    def start(self):
        """starts reading and logging"""
        program = self.extra.get("program", "")
        name = f"{program}:{self.log_level}"
        thread = Thread(target=lambda: self._run(), name=name)
        thread.daemon = True
        self.thread = thread
        thread.start()
        return self

    def close(self):
        try:
            self.stream.flush()
        except IOError:
            pass
        self.stream.close()


class ExecutionContext(object):
    """
    The context in which a program is executed

    It contains all non-static resources necessary to execute a program,
    such as the logger and the system environment.
    """

    def __init__(self, environment: Dict[str, str], observer: ProgramObserver) -> None:
        self.environment = environment
        self.observer = observer
        extensions.extend_environment(observer.get_name(), environment)

    def extend(self, environment: Dict[str, str]) -> "ExecutionContext":
        env = deepcopy(self.environment)
        if environment:
            env.update(environment)
            extensions.extend_environment(self.observer.get_name(), self.environment)

        return ExecutionContext(env, self.observer)

    def spawn(
        self,
        name: str,
        environment: Dict[str, str],
        logger: Logger,
        extra: Dict[str, str],
    ) -> "ExecutionContext":
        env = deepcopy(self.environment)
        observer = self.observer.spawn(name, logger, extra)

        if environment:
            extensions.extend_environment(name, self.environment)
            env.update(environment)

        extensions.update_logger(name, logger)
        return ExecutionContext(env, observer)


class Program(object):
    """
    Starts a program and caputures stdout and sterr as log messages
    """

    def __init__(
        self, name: str, config: ProgramConfig, parent_context: ExecutionContext
    ) -> None:
        """
        :name: (str)
            the program name
        :config: (ProgramConfig)
            the program configuration
        :parent_context (ExecutionContext):
            the parent program execution context
        """
        self.name = name
        self.config = config

        self.logger = getLogger(__name__)

        if config.loglevel:
            self.logger.setLevel(config.loglevel)

        self.extra = {"program": name}
        self.context: ExecutionContext = parent_context.spawn(
            name, self.config.environment or {}, self.logger, self.extra
        )

        if self.config.command:
            self.shell = False
            assert isinstance(self.config.command, list)
            self.command = self.config.command
        else:
            self.shell = True
            assert isinstance(self.config.sh, str)
            self.command = [self.config.sh]

        observer = self.context.observer

        self._observer = observer
        self._state_handler = ProgramStateHandler(observer)
        self._process: Optional[Popen] = None

    def _run(self) -> None:
        logger = self.logger
        extra = self.extra
        command = self.command
        env = self.context.environment
        state = self._state_handler
        observer = self._observer
        startup_delay = self.config.startup_delay
        umask = self.config.umask
        out: Optional[LogStream] = None
        err: Optional[LogStream] = None

        try:
            assert isinstance(startup_delay, float) or isinstance(startup_delay, int)
            assert isinstance(umask, int)

            state.wait(float(startup_delay))

            observer.on_execution(command, env, self.config)
            state.set(ProgramState.STARTING)

            shell = self.shell
            args: Union[str, List[str]] = command

            if shell:
                args = command[0]

            process = Popen(
                args,
                stdout=PIPE,
                stderr=PIPE,
                env=env,
                user=self.config.user,
                umask=umask,
                shell=shell,
                start_new_session=True,
            )
            assert process.stdout is not None
            assert process.stderr is not None

            state.set(ProgramState.RUNNING)
            self._process = process

            observer.on_run(process.pid)

            err = LogStream(logger, ERROR, process.stderr, extra).start()
            out = LogStream(logger, INFO, process.stdout, extra).start()

            process.wait()
            state.handle_exit(process.returncode, self.command)
        except ProgramCanceledException:
            observer.on_cancel()
            state.set(ProgramState.CANCELED)
        except BaseException as e:
            observer.on_crash(self.command, e)
            state.set(ProgramState.CRASHED)
        finally:
            self._process = None
            if out:
                out.close()
            if err:
                err.close()

    def get_state(self) -> int:
        return self._state_handler.get()

    def start(self, timeout: Optional[float] = 1) -> int:
        thread = Thread(target=lambda: self._run(), name=self.name)
        thread.daemon = True
        thread.start()

        self._observer.on_start()

        return self._state_handler.wait_for_startup(timeout)

    def join(
        self,
        timeout: Optional[float] = None,
    ) -> ProgramState:
        return self._state_handler.join(timeout)

    def interrupt(self) -> None:
        return self._state_handler.kill(self._process, SIGINT)

    def terminate(self) -> None:
        return self._state_handler.kill(self._process, SIGTERM)  # type: ignore

    def join_wait(self, timeout: Optional[float] = None) -> ProgramState:
        return self._state_handler.join_wait(timeout)
